// Use an oszilloscope as display

// $Date: 2024-12-04 13:21:56 +0100 (Mi, 04. Dez 2024) $

// for brownout detection
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include <driver/dac_oneshot.h>

#include "wifitime.h"

struct {
  bool minute_ticks;
  bool xmas;
} Config = {
  false, true
};

// digital to analog channel handles
dac_oneshot_handle_t DAC[2];

// internal time format 4 digit HHMM
int getLocalTime_HHMM() {
  struct tm timeinfo;
  int t = -1;
  if (getLocalTime(&timeinfo)) {
    int h = timeinfo.tm_hour;
    t = h * 100 + timeinfo.tm_min;
  }
  return t;
}

#include <ArduinoGraphics.h>

class OSZIGraphics : public ArduinoGraphics {
  uint8_t xy[512][2];  // the display ... the dots to display
public:
  int xy_last = 0;

  OSZIGraphics()
    : ArduinoGraphics(256, 256) {
  }

  void dump() {
    for (int i = 0; i < xy_last; i++) {
      Serial.print(xy[i][0]);
      Serial.print(" ");
      Serial.println(xy[i][1]);
    }
  }

  void set(int x, int y, uint8_t r, uint8_t g, uint8_t b) {
    xy[xy_last][0] = x;
    xy[xy_last][1] = y;
    xy_last++;
  }

  // tick must be called in loop or an isr
  // returns true if new screen starts.
  bool tick() {
    static int xy_index = 0;
    dac_oneshot_output_voltage(DAC[0], xy[xy_index][0]);
    dac_oneshot_output_voltage(DAC[1], xy[xy_index][1]);
    xy_index++;
    if (xy_index > xy_last) {
      xy_index = 0;
      return true;
    }
    return false;
  }

  void line(int x0, int y0, int x1, int y1) {
    // do not draw horizontal from higher to lower x
    if ((x0 > x1) && (y0 == y1)) {
      int x = x0;
      x0 = x1;
      x1 = x;
    }
    // do not draw vertical from higher to lower x
    if ((x0 == x1) && (y0 > y1)) {
      int y = y0;
      y0 = y1;
      y1 = y;
    }
    ArduinoGraphics::line(x0, y0, x1, y1);
  }
};

OSZIGraphics oszi = OSZIGraphics();

class AnalogClock {
public:
  // index after clock face
  int xy_index_eof = 0;

  int mid[2] = { 100, 100 };
  int watch_rad = 50;
  OSZIGraphics* _oszi;
  AnalogClock(OSZIGraphics* disp) {
    _oszi = disp;
  }

  void drawFace() {
    _oszi->stroke(127, 127, 127);  // color
    if (Config.xmas) {
      // draw the tree
      int lines[] = {
        // x y
        30, 10, 32, 20,
        32, 20, 15, 18,
        16, 18, 34, 34,
        34, 34, 25, 34,
        25, 34, 38, 63, // tip
        38, 63, 48, 44,
        48, 44, 42, 45,
        42, 45, 55, 30,
        55, 30, 47, 31,
        47, 31, 64, 15,
        64, 15, 40, 19,
        40, 19, 43, 10,
        43, 10, 31, 9,
        // star
        50,82,54,92, 
        55,90,42,97,  // tip2
        42,97,56,99,
        56,99,60,109, // tip 3
        60,109,63,100,
        64,100,77,98,  // tip 4
        77,98,64,90,
        64,90,64,80,  // tip 5
        64,80,59,88,
        59,88,50,82   // tip 1
      };
      int MAX_LINE = sizeof(lines) / sizeof(lines[0]);
      for (int i = 0; i < MAX_LINE; i += 4) {
        _oszi->line(lines[i],
                    lines[i + 1],
                    lines[i + 2],
                    lines[i + 3]);
      }
      // shift the watch and scale down
      mid[0] = 125;
      mid[1] = 55;
      watch_rad = 25;
    }
    for (int i = 0; i < 12; i++) {
      float _angle = PI * 2 * i / 12.;
      _oszi->line(mid[0] + watch_rad * cos(_angle),
                  mid[1] + watch_rad * sin(_angle),
                  mid[0] + (watch_rad + 7) * cos(_angle),
                  mid[1] + (watch_rad + 7) * sin(_angle));
      if (Config.minute_ticks) {
        int _sub = 10;
        for (int j = 0; j < _sub; j++) {
          float _angle = PI * 2 * (i + (j / float(_sub))) / float(_sub);
          _oszi->line(mid[0] + (watch_rad + 7) * cos(_angle),
                      mid[1] + (watch_rad + 7) * sin(_angle),
                      mid[0] + (watch_rad + 7) * cos(_angle),
                      mid[1] + (watch_rad + 7) * sin(_angle));
        }
      }
    }
    xy_index_eof = _oszi->xy_last;
  }

  void drawHand(int len, float angle) {
    int cos_ = len * cos(angle);
    int sin_ = len * sin(angle);
    // BUG line is direction dependend
    _oszi->line(mid[0], mid[1], mid[0] + cos_, mid[1] + sin_);
  }

  void drawHands(int number) {
    // hhmm
    int mm = number % 100;
    float hh = int(number / 100) + float(mm) / 60;
    _oszi->xy_last = xy_index_eof;

    float mm_angle = PI / 2 + PI * 2 / 60. * (60 - mm);
    drawHand(watch_rad, mm_angle);

    float hh_angle = PI / 2 + PI * 2 / 12. * (12 - hh);
    drawHand(watch_rad - 10, hh_angle);
  }

  void animation() {
    static int angle = 0;
    _oszi->xy_last = xy_index_eof;
    // snowfall
    drawHand(30, angle);
    angle -= 1;
  }

  void dump() {
    _oszi->dump();
  }
};

AnalogClock aClock = AnalogClock(&oszi);

void setup() {
  Serial.begin(115200);
  // brownout detection off ... at least for wifi connect, when on PC USB
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);  //disable brownout detector

  Serial.println("ESP32 Osziclock using DAC1 and DAC2, pin 25 and 26.");
  if (!oszi.begin()) {
    Serial.println("Failed to initialize the display!");
    // impossible
  }
  Serial.print("Draw clock face:");
  aClock.drawFace();
  Serial.print(aClock.xy_index_eof);
  Serial.println("  points");

  Serial.println("\nWiFi setup");
  wifiSetup();
  dac_oneshot_config_t dacfg;
  dacfg.chan_id = DAC_CHAN_0;
  dac_oneshot_new_channel(&dacfg, &DAC[0]);
  dacfg.chan_id = DAC_CHAN_1;
  dac_oneshot_new_channel(&dacfg, &DAC[1]);
}

void loop() {
  static bool fetch_time = true;
  // TODO refetch_time
  if (fetch_time) {
    if (getTime_from_network()) {
      fetch_time = false;
    }
  }
  if (oszi.tick()) {
    // one page for a watch is 10ms
    static int cnt = 32000;
    if (cnt > 1000) {  // do not fetch time always
      cnt = 0;
      if (fetch_time) {
        aClock.animation();
      } else {
        static int time_ = 0;
        int t = getLocalTime_HHMM();
        if (t != time_) {
          time_ = t;
          Serial.print("localtime ");
          Serial.println(time_);
          aClock.drawHands(time_);
        }
      }
    }
    cnt++;
  }
}